import numpy as np
import dolfin as dl
import warnings
from numbers import Number
from cuqi.samples import Samples

def _import_ufl():
    """Import ufl module. This function is used to make importing ufl compatible
    with FEniCS versions that name `ufl` module as`ufl` or alternatively as 
    `ufl_legacy`.
    """
    try:
        import ufl_legacy as ufl
    except (Exception, RuntimeError, ImportError):
        import ufl
    return ufl

def _compute_stats(samples: Samples):
    """This function computes the statistics (mean and variance) of a set of 
    samples on the function value representation. Two different approaches 
    used in computing the variance: The first approach is computing the variance
    var1 using the FEniCS functions directly, and is general for any FEM 
    function. The second approach is computing the variance var2 using the
    FEniCS vectors directly, and is specific for some FEM function spaces,
    e.g. CG1. The returned values are the mean, var1, and var2.

    Note that the second approach gives the same result as what is
    obtained by the command `samples.funvals.vector.variance()`
    """

    geom = samples.geometry
    # raise error if geom does not have a function_space attribute
    if not hasattr(geom, 'function_space'):
        raise AttributeError("The geometry must have a "+\
                             "function_space attribute")
 
    V = geom.function_space

    # Loop to compute the samples function value
    Ns = samples.samples.shape[-1] # Number of samples
    sample_funs = []
    sample_funs_dof_vecs = np.empty((V.dim(), Ns)) 
    for i, sample in enumerate(samples):
        sample_funs.append(geom.par2fun(sample))
        sample_funs_dof_vecs[:, i] = sample_funs[-1].vector().get_local()

    # Sample mean
    sample_mean_dof = np.mean(sample_funs_dof_vecs, axis=1)
    sample_mean_f = dl.Function(V)
    sample_mean_f.vector().set_local(sample_mean_dof)

    # Compute variance 
    # Approach 1 (this approach is general for all FEM function spaces)

    # Create a function space with higher order elements
    V2 = dl.FunctionSpace(V.mesh(), V.ufl_element().family(), V.ufl_element().degree()*2)

    # Loop to create terms required for variance computation
    var_terms = np.empty((V2.dim(), Ns))
    for i, sample_f in enumerate(sample_funs):
        expr_f = dl.project(sample_f*sample_f - 2*sample_mean_f*sample_f, V2)
        var_terms[:, i] = expr_f.vector().get_local()
    
    mean_var_terms = np.mean(var_terms, axis=1)
    mean_var_terms_f = dl.Function(V2)
    mean_var_terms_f.vector().set_local(mean_var_terms)
    
    var1 = dl.project(mean_var_terms_f + sample_mean_f*sample_mean_f , V2)

    # Approach 2 (this approach is specific for some FEM function spaces, e.g. CG1)
    if V.ufl_element().family() != 'Lagrange':
        warnings.warn("The function space is not Lagrange, the variance, var2,"+ 
                     "computed using the second approach may not be correct.")
    var2_vec = np.var(sample_funs_dof_vecs, axis=1)
    var2 = dl.Function(V)
    var2.vector().set_local(var2_vec)

    return sample_mean_f, var1, var2


class ExpressionFromCallable(dl.UserExpression):
    """A `dolfin.Expression` that is created from a callable. The callable
    should take a single argument, the coordinate vector `x` and return a
    number, the value of the expression at `x`. The keyword arguments are passed
    to the `dolfin.Expression` constructor. An example of the keyword arguments 
    is the `degree` argument for `dolfin.Expression` which is used to specify 
    the element (polynomial) degree of the expression.
    """
    def __init__(self, func, **kwargs):
        super().__init__(**kwargs)
        self.func = func

    def eval(self, value, x):
        value[:] = self.func(x)


def to_dolfin_expression(value, **expression_kwargs):
    """Converts a value to a `dolfin.Expression`. The value should be a 
    callable, `dolfin.Expression` or a number. expression_kwargs are passed to
    the `dolfin.Expression` constructor if the value is a callable. An example 
    of the keyword arguments is the `degree` argument for `dolfin.Expression` 
    which is used to specify the element (polynomial) degree of the expression.
    """
    
    if not callable(value) and expression_kwargs != {}:
        raise ValueError("Cannot pass kwargs to non-callable value")
    # Note that dl.Expression and dl.Constant objects are
    # callable objects, so we need to check for them first
    if isinstance(value, dl.UserExpression) \
        or isinstance(value, dl.Expression)\
            or isinstance(value, dl.Constant):
        return value
    elif callable(value):
        return ExpressionFromCallable(value, **expression_kwargs)
    elif isinstance(value, Number):
        return dl.Constant(value)
    else:
        raise ValueError("Cannot convert to dolfin.Expression")



import pytest
from cuqipy_fenics.geometry import FEniCSContinuous, FEniCSMappedGeometry, MaternKLExpansion
from cuqi.distribution import Gaussian
from cuqipy_fenics.utilities import _compute_stats
import dolfin as dl
import numpy as np
import matplotlib.pyplot as plt


@pytest.fixture(params=["G_FE", "G_KL", "G_map"])
def samples(request):
    """ Create Samples object with different geometries"""
    # Define the mesh and the function space
    geometry_type = request.param
    Ns = 5
    mesh = dl.UnitSquareMesh(10, 10)
    V = dl.FunctionSpace(mesh, "CG", 2) 

    # Define the geometries depending on the case
    # FEniCSContinuous geometry
    geom = FEniCSContinuous(V)
    
    if geometry_type == "G_KL" or geometry_type == "G_map":
        # A KL expansion geometry
        num_terms = 10
        length_scale = 0.05
        geom = MaternKLExpansion(geom, length_scale, num_terms)
    
    if geometry_type == "G_map":
        # A mapped geometry applied to the KL expansion geometry
        c_minus = 1
        c_plus = 10
        def heavy_map(func):
            dofs = func.vector().get_local()
            updated_dofs = c_minus*0.5*(1 + np.sign(dofs)) +\
                c_plus*0.5*(1 - np.sign(dofs))
            func.vector().set_local(updated_dofs)
            return func
        geom = FEniCSMappedGeometry(geom, heavy_map)
    
    # Create a distribution and sample the distribution
    x = Gaussian(0, np.ones(geom.par_dim), geometry=geom)
    samples = x.sample(Ns)
    return samples


def test_computing_samples_statistics_is_correct(samples):
    """Test that the mean and variance computed on the samples'
    function representation and the samples' parameter representation
    are correct."""
    
    V = samples.geometry.function_space
    samples_funvals = samples.funvals

    # Compute mean and variance on the **parameter** space
    mean = samples.mean()
    var = samples.variance()

    assert np.allclose(mean, samples.samples.mean(axis=1))
    assert np.allclose(var, samples.samples.var(axis=1))

    # Compute mean and variance on the **function** space (this will generate an
    # error because the samples are not converted to vector form)
    with pytest.raises(Exception, match="cuqi.samples._samples added message:"):
        samples_funvals.mean()

    with pytest.raises(Exception, match="cuqi.samples._samples added message:"):
        samples_funvals.variance()

    # Compute mean and variance on the **function** space after converting the
    # samples to vector form
    mean_funvals = samples_funvals.vector.mean()
    var_funvals = samples_funvals.vector.variance()

    # Compute mean and variance on the **function** space using helper function
    mean_helper, var1_helper, var2_helper = _compute_stats(samples)

    # var1_helper is computed on a higher order function space, so we need to
    # evaluate it at the DOFs of the original function space V to compare with
    # var_funvals
    var1_helper_V_DOF = np.empty(V.dim())
    for i in range(V.dim()):
        var1_helper_V_DOF[i] = var1_helper(V.tabulate_dof_coordinates()[i,:])
 
    # Assert that the mean and variance computed on the function space are 
    # correct
    assert np.allclose(mean_funvals, mean_helper.vector().get_local())
    assert np.allclose(var_funvals, var2_helper.vector().get_local())
    assert np.allclose(var_funvals, var1_helper_V_DOF)


@pytest.mark.parametrize("mesh", [dl.UnitSquareMesh(20, 20), dl.UnitIntervalMesh(20)])  
@pytest.mark.parametrize("basis", ['CG', 'DG'])
@pytest.mark.parametrize("order", [1, 2])
def test_method_plot_envelope_works_only_for_1D_1st_order_Lagrange_space(
        mesh, basis, order):
    """Test that the method plot_envelope used in plot_ci works only for 
    1D 1st order Lagrange space and raises an error otherwise"""

    # define the function space
    space = dl.FunctionSpace(mesh, basis, order)

    # Set the geometry (mapped KL) 
    G_FEM = FEniCSContinuous(space)
    G_KL = MaternKLExpansion(G_FEM, length_scale=0.1, nu=0.75, num_terms=10)

    # define some map
    def prior_map(func):
        dofs = func.vector().get_local()
        updated_dofs = 15*dofs
        func.vector().set_local(updated_dofs)
        return func

    G = FEniCSMappedGeometry(G_KL, map=prior_map)

    # define distribution
    x = Gaussian(0, cov=1, geometry=G)

    # sample the distribution
    samples = x.sample(20)
    exact = x.sample()

    # plot the envelope only if the order is 1 and the basis is CG
    # and the physical dimension is 1. Otherwise, assert that the method
    # plot_envelope raises an error
    if order == 1 and basis == 'CG' and G_FEM.physical_dim == 1:
        # plot_ci which uses _plot_envelope method
        samples.funvals.vector.plot_ci(exact=exact)
    else:
        # assert that the method plot_envelope raises an error
        with pytest.raises(NotImplementedError,
                           match="Envelope plot is only implemented"):
            samples.funvals.vector.plot_ci(exact=exact)